
/* Chrysalide - Outil d'analyse de fichiers binaires
 * instruction.c - gestion des instructions de la ARM
 *
 * Copyright (C) 2013 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  OpenIDA is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  OpenIDA is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "instruction.h"


#include <string.h>


#include "instruction-int.h"
#include "../../common/extstr.h"



/* Initialise la classe des instructions ARM. */
static void g_arm_instruction_class_init(GArmInstructionClass *);

/* Initialise une instance de instruction ARM. */
static void g_arm_instruction_init(GArmInstruction *);

/* Supprime toutes les références externes. */
static void g_arm_instruction_dispose(GArmInstruction *);

/* Procède à la libération totale de la mémoire. */
static void g_arm_instruction_finalize(GArmInstruction *);

/* Reconstruit le cache complet d'une désignation d'instruction. */
static void g_arm_instruction_build_keyword(const GArmInstruction *, AsmSyntax);



/* Indique le type défini pour une représentation d'une instruction ARM. */
G_DEFINE_TYPE(GArmInstruction, g_arm_instruction, G_TYPE_ARCH_INSTRUCTION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des instructions ARM.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_class_init(GArmInstructionClass *klass)
{
    GObjectClass *object_class;             /* Autre version de la classe  */
    GArchInstructionClass *instr;           /* Encore une autre vision...  */

    object_class = G_OBJECT_CLASS(klass);
    instr = G_ARCH_INSTRUCTION_CLASS(klass);

    object_class->dispose = (GObjectFinalizeFunc/* ! */)g_arm_instruction_dispose;
    object_class->finalize = (GObjectFinalizeFunc)g_arm_instruction_finalize;

    instr->build_key = (build_instruction_keyword_fc)g_arm_instruction_build_keyword;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance à initialiser.                              *
*                                                                             *
*  Description : Initialise une instance d'instruction ARM.                   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_init(GArmInstruction *instr)
{
    instr->cond = ACC_AL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_dispose(GArmInstruction *instr)
{
    G_OBJECT_CLASS(g_arm_instruction_parent_class)->dispose(G_OBJECT(instr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_finalize(GArmInstruction *instr)
{
    G_OBJECT_CLASS(g_arm_instruction_parent_class)->finalize(G_OBJECT(instr));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr  = instruction à traiter.                              *
*                syntax = type de représentation demandée.                    *
*                                                                             *
*  Description : Reconstruit le cache complet d'une désignation d'instruction.*
*                                                                             *
*  Retour      : Mot clef de bas niveau.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arm_instruction_build_keyword(const GArmInstruction *instr, AsmSyntax syntax)
{
    GArchInstruction *base;                 /* Instruction, vue générique  */

    base = G_ARCH_INSTRUCTION(instr);

    base->cached_keyword = strdup(instr->keyword);

    if (base->suffix != NULL)
        base->cached_keyword = stradd(base->cached_keyword, base->suffix);

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instruction ARM à mettre à jour.                     *
*                cond  = condition d'exécution de l'instruction.              *
*                                                                             *
*  Description : Définit les conditions d'exécution d'une instruction ARM.    *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_arm_instruction_set_cond(GArmInstruction *instr, ArmCondCode cond)
{
    instr->cond = cond;

    return true;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : instr = instruction ARM à consulter.                         *
*                                                                             *
*  Description : Indique les conditions d'exécution d'une instruction ARM.    *
*                                                                             *
*  Retour      : Condition d'exécution de l'instruction.                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

ArmCondCode g_arm_instruction_get_cond(const GArmInstruction *instr)
{
    return instr->cond;

}
